# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2011 Carlos Abalde
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Google Documents List API backend.
"""

import os.path
import string
import urllib.error
import urllib.parse
import urllib.request

import duplicity.backend
from duplicity import __version__
from duplicity.errors import BackendException


class GDocsBackend(duplicity.backend.Backend):
    """
    Connect to remote store using Google Google Documents List API
    """

    ROOT_FOLDER_ID = "folder%3Aroot"
    BACKUP_DOCUMENT_TYPE = "application/binary"

    def __init__(self, parsed_url):
        """
        Initialize the GDocs backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Import Google Data APIs libraries.
        try:
            global atom
            global gdata
            import atom.data
            import gdata.client
            import gdata.docs.client
            import gdata.docs.data
        except ImportError as e:
            raise BackendException(
                f"Google Docs backend requires Google Data APIs Python Client Library\n"
                f"(see http://code.google.com/p/gdata-python-client/).\n"
                f"Exception: {str(e)}"
            )

        # Setup client instance.
        self.client = gdata.docs.client.DocsClient(source=f"duplicity {__version__}")
        self.client.ssl = True
        self.client.http_client.debug = False
        self._authorize(f"{parsed_url.username}@{parsed_url.hostname}", self.get_password())

        # Fetch destination folder entry (and crete hierarchy if required).
        folder_names = string.split(parsed_url.path[1:], "/")
        parent_folder = None
        parent_folder_id = GDocsBackend.ROOT_FOLDER_ID
        for folder_name in folder_names:
            entries = self._fetch_entries(parent_folder_id, "folder", folder_name)
            if entries is not None:
                if len(entries) == 1:
                    parent_folder = entries[0]
                elif len(entries) == 0:
                    folder = gdata.docs.data.Resource(type="folder", title=folder_name)
                    parent_folder = self.client.create_resource(folder, collection=parent_folder)
                else:
                    parent_folder = None
                if parent_folder:
                    parent_folder_id = parent_folder.resource_id.text
                else:
                    raise BackendException(f"Error while creating destination folder '{folder_name}'.")
            else:
                raise BackendException(f"Error while fetching destination folder '{folder_name}'.")
        self.folder = parent_folder

    def _put(self, source_path, remote_filename):
        self._delete(remote_filename)

        # Set uploader instance. Note that resumable uploads are required in order to
        # enable uploads for all file types.
        # (see http://googleappsdeveloper.blogspot.com/2011/05/upload-all-file-types-to-any-google.html)
        file = source_path.open()
        uploader = gdata.client.ResumableUploader(
            self.client,
            file,
            GDocsBackend.BACKUP_DOCUMENT_TYPE,
            os.path.getsize(file.name),
            chunk_size=gdata.client.ResumableUploader.DEFAULT_CHUNK_SIZE,
            desired_class=gdata.docs.data.Resource,
        )
        if uploader:
            # Chunked upload.
            entry = gdata.docs.data.Resource(title=atom.data.Title(text=remote_filename))
            uri = f"{self.folder.get_resumable_create_media_link().href}?convert=false"
            entry = uploader.UploadFile(uri, entry=entry)
            if not entry:
                raise BackendException(
                    f"Failed to upload file '{source_path.get_filename()}' "
                    f"to remote folder '{self.folder.title.text}'"
                )
        else:
            raise BackendException(
                f"Failed to initialize upload of file '{source_path.get_filename()}' "
                f"to remote folder '{self.folder.title.text}'"
            )
        assert not file.close(), "file failed to close"

    def _get(self, remote_filename, local_path):
        entries = self._fetch_entries(
            self.folder.resource_id.text,
            GDocsBackend.BACKUP_DOCUMENT_TYPE,
            remote_filename,
        )
        if len(entries) == 1:
            entry = entries[0]
            self.client.DownloadResource(entry, local_path.name)
        else:
            raise BackendException(
                f"Failed to find file '{remote_filename}' " f"in remote folder '{self.folder.title.text}'"
            )

    def _list(self):
        entries = self._fetch_entries(self.folder.resource_id.text, GDocsBackend.BACKUP_DOCUMENT_TYPE)
        return [entry.title.text for entry in entries]

    def _delete(self, filename):
        entries = self._fetch_entries(self.folder.resource_id.text, GDocsBackend.BACKUP_DOCUMENT_TYPE, filename)
        for entry in entries:
            self.client.delete(f"{entry.get_edit_link().href}?delete=true", force=True)

    def _authorize(self, email, password, captcha_token=None, captcha_response=None):
        try:
            self.client.client_login(
                email,
                password,
                source=f"duplicity {__version__}",
                service="writely",
                captcha_token=captcha_token,
                captcha_response=captcha_response,
            )
        except gdata.client.CaptchaChallenge as challenge:
            print(f"A captcha challenge in required. Please visit {challenge.captcha_url}")
            answer = None
            while not answer:
                answer = eval(input("Answer to the challenge? "))
            self._authorize(email, password, challenge.captcha_token, answer)
        except gdata.client.BadAuthentication:
            raise BackendException(
                "Invalid user credentials given. Be aware that accounts "
                "that use 2-step verification require creating an application specific "
                "access code for using this Duplicity backend. Follow the instruction in "
                "http://www.google.com/support/accounts/bin/static.py?page=guide.cs&guide=1056283&topic=1056286 "
                "and create your application-specific password to run duplicity backups."
            )

    def _fetch_entries(self, folder_id, type, title=None):  # pylint: disable=redefined-builtin
        # Build URI.
        uri = f"/feeds/default/private/full/{folder_id}/contents"
        if type == "folder":
            uri += "/-/folder?showfolders=true"
        elif type == GDocsBackend.BACKUP_DOCUMENT_TYPE:
            uri += "?showfolders=false"
        else:
            uri += "?showfolders=true"
        if title:
            uri += f"&title={urllib.parse.quote(title)}&title-exact=true"

        # Fetch entries.
        entries = self.client.get_all_resources(uri=uri)

        # When filtering by entry title, API is returning (don't know why) documents in other
        # folders (apart from folder_id) matching the title, so some extra filtering is required.
        if title:
            result = []
            for entry in entries:
                resource_type = entry.get_resource_type()
                if (
                    (not type)
                    or (type == "folder" and resource_type == "folder")
                    or (type == GDocsBackend.BACKUP_DOCUMENT_TYPE and resource_type != "folder")
                ):
                    if folder_id != GDocsBackend.ROOT_FOLDER_ID:
                        for link in entry.in_collections():
                            folder_entry = self.client.get_entry(
                                link.href,
                                None,
                                None,
                                desired_class=gdata.docs.data.Resource,
                            )
                            if folder_entry and (folder_entry.resource_id.text == folder_id):
                                result.append(entry)
                    elif len(entry.in_collections()) == 0:
                        result.append(entry)
        else:
            result = entries

        # Done!
        return result


"""
gdata is an alternate way to access gdocs, currently 05/2015 lacking OAuth support.
"""
duplicity.backend.register_backend("gdata+gdocs", GDocsBackend)
duplicity.backend.uses_netloc.extend(["gdata+gdocs"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Manage temporary files
"""

import os
import shutil
import sys

from duplicity import config
from duplicity import file_naming
from duplicity import gpg
from duplicity import errors
from duplicity import log
from duplicity import path
from duplicity import tempdir


def new_temppath():
    """
    Return a new TempPath
    """
    filename = tempdir.default().mktemp()
    return TempPath(filename)


class TempPath(path.Path):
    """
    Path object used as a temporary file
    """

    def delete(self):
        """
        Forget and delete
        """
        path.Path.delete(self)
        tempdir.default().forget(self.name)

    def open_with_delete(self, mode):
        """
        Returns a fileobj.  When that is closed, delete file
        """
        fh = FileobjHooked(path.Path.open(self, mode))
        fh.addhook(self.delete)
        return fh


def get_fileobj_duppath(dirpath, partname, permname, remname, overwrite=False):
    """
    Return a file object open for writing, will write to the given filename.

    Data will be processed and written to a temporary file. When the
    returned file object is closed, it is renamed to its final position.
    The filename must be a recognizable duplicity data file.

    :param Path dirpath: The directory path.
    :param str partname: The partial filename.
    :param str permname: The permanent filename.
    :param str remname: The remote filename.
    :param bool overwrite: Whether to overwrite existing files.
    :returns: A hooked file object.
    :rtype: FileobjHooked
    """
    if not config.restart:
        td = tempdir.TemporaryDirectory(dirpath.name)
        tdpname = td.mktemp()
        tdp = TempDupPath(tdpname, parseresults=file_naming.parse(partname))
        fh = FileobjHooked(
            tdp.filtered_open("wb"),
            tdp=tdp,
            dirpath=dirpath,
            partname=partname,
            permname=permname,
            remname=remname,
        )
    else:
        dp = path.DupPath(dirpath.name, index=(partname,))
        mode = "ab"
        if overwrite:
            mode = "wb"
        fh = FileobjHooked(
            dp.filtered_open(mode),
            tdp=None,
            dirpath=dirpath,
            partname=partname,
            permname=permname,
            remname=remname,
        )

    def rename_and_forget():
        tdp.rename(dirpath.append(partname))
        td.forget(tdpname)

    if not config.restart:
        fh.addhook(rename_and_forget)

    return fh


def new_tempduppath(parseresults):
    """
    Return a new ``TempDupPath``, using settings from ``parseresults``.

    :param ParseResults parseresults: The results of parsing a filename.
    :returns: A new TempDupPath instance.
    :rtype: TempDupPath
    """
    filename = tempdir.default().mktemp()
    return TempDupPath(filename, parseresults=parseresults)


class TempDupPath(path.DupPath):
    """
    Like TempPath, but build around DupPath
    """

    def delete(self):
        """
        Forget and delete
        """
        path.DupPath.delete(self)
        tempdir.default().forget(self.name)

    def filtered_open_with_delete(self, mode):
        """
        Returns a filtered fileobj.  When that is closed, delete file
        """
        fh = FileobjHooked(path.DupPath.filtered_open(self, mode))
        fh.addhook(self.delete)
        return fh

    def open_with_delete(self, mode="rb"):
        """
        Returns a fileobj.  When that is closed, delete file
        """
        assert mode == "rb", "wrong mode for open_with_delete"
        fh = FileobjHooked(path.DupPath.open(self, mode))
        fh.addhook(self.delete)
        return fh


class FileobjHooked(object):
    """
    Simulate a file, but add hook on close
    """

    def __init__(
        self,
        fileobj,
        tdp=None,
        dirpath=None,
        partname=None,
        permname=None,
        remname=None,
    ):
        """
        Initializer.  fileobj is the file object to simulate
        """
        self.fileobj = fileobj  # the actual file object
        self.closed = False  # True if closed
        self.hooklist = []  # filled later with thunks to run on close
        self.tdp = tdp  # TempDupPath object
        self.dirpath = dirpath  # path to directory
        self.partname = partname  # partial filename
        self.permname = permname  # permanent filename
        self.remname = remname  # remote filename

    def write(self, buf):
        """
        Write fileobj, return result of write()
        """
        return self.fileobj.write(buf)

    def flush(self):
        """
        Flush fileobj and force sync.
        """
        self.fileobj.flush()
        os.fsync(self.fileobj.fileno())

    def to_partial(self):
        """
        Rename tdp to partname and remove the rename_and_forget hook
        """
        log.Notice(f"DEBUG: to_partial called for {self.partname}")
        assert not config.restart, "may not restart after checkpoint"
        self.fileobj.flush()
        self.tdp.rename(self.dirpath.append(self.partname))
        del self.hooklist[0]

    def to_remote(self):
        """
        We have written the last checkpoint, now encrypt or compress
        and send a copy of it to the remote for final storage.
        """
        log.Notice(f"DEBUG: to_remote called for {self.remname}")
        pr = file_naming.parse(self.remname)
        src = self.dirpath.append(self.partname)
        tgt = self.dirpath.append(self.remname)
        src_iter = SrcIter(src)
        if pr.compressed:
            gpg.GzipWriteFile(src_iter, tgt.name, size=sys.maxsize)
        elif pr.encrypted:
            gpg.GPGWriteFile(src_iter, tgt.name, config.gpg_profile, size=sys.maxsize)
        else:
            shutil.copyfile(src.name, tgt.name)
        config.backend.move(tgt)

    def to_final(self):
        """
        We are finished, rename to final, gzip if needed.
        """
        log.Notice(f"DEBUG: to_final called for {self.permname}")
        src = self.dirpath.append(self.partname)
        tgt = self.dirpath.append(self.permname)
        src_iter = SrcIter(src)
        pr = file_naming.parse(self.permname)
        if pr.compressed:
            gpg.GzipWriteFile(src_iter, tgt.name, size=sys.maxsize)
            os.unlink(src.name)
        else:
            os.rename(src.name, tgt.name)

    def read(self, length=-1):
        """
        Read fileobj, return result of read()
        """
        return self.fileobj.read(length)

    def tell(self):
        """
        Returns current location of fileobj
        """
        return self.fileobj.tell()

    def seek(self, offset):
        """
        Seeks to a location of fileobj
        """
        return self.fileobj.seek(offset)

    def close(self):
        """
        Close fileobj, running hooks right afterwards
        """
        assert not self.fileobj.close(), "self.fileobj failed to close"
        for hook in self.hooklist:
            hook()

    def clean_up(self):
        """
        delete partial files
        """
        self.dirpath.append(self.partname).delete()

    def addhook(self, hook):
        """
        Add hook (function taking no arguments) to run upon closing
        """
        self.hooklist.append(hook)

    def get_name(self):
        """
        Return the name of the file
        """
        return self.fileobj.name

    name = property(get_name)


class Block(object):
    """
    Data block to return from SrcIter
    """

    def __init__(self, data):
        self.data = data


class SrcIter(object):
    """
    Iterate over source and return Block of data.
    """

    def __init__(self, src):
        self.src = src
        self.fp = src.open("rb")

    def __next__(self):
        try:
            res = Block(self.fp.read(self.get_read_size()))
        except Exception:
            raise errors.FatalError(
                _("Failed to read %s: %s") % (self.src.uc_name, sys.exc_info()),
                log.ErrorCode.generic,
            )
        if not res.data:
            self.fp.close()
            raise StopIteration
        return res

    def get_read_size(self):
        return 128 * 1024

    def get_footer(self):
        return b""



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


"""
Functions for patching of directories.
"""

import os
import re
import tempfile

from duplicity import diffdir
from duplicity import dup_tarfile
from duplicity import errors
from duplicity import librsync
from duplicity import log
from duplicity import selection
from duplicity import tempdir
from duplicity import util
from duplicity.lazy import ITRBranch
from duplicity.lazy import IterTreeReducer
from duplicity.path import ROPath


class PatchDirException(Exception):
    """
    Base exception for directory patching errors.
    """

    pass


def filter_path_iter(path_iter, index):
    """
    Rewrite path elements of path_iter so they start with index.

    Discard any that doesn't start with index, and remove the index
    prefix from the rest.

    :param iterator path_iter: Iterator of ROPaths.
    :param tuple index: The path prefix to filter and strip.
    :returns: An iterator of rewritten ROPaths.
    :rtype: iterator
    """
    assert isinstance(index, tuple) and index, "Index must be a non-empty tuple specifying the path prefix"
    l = len(index)
    for path in path_iter:
        if path.index[:l] == index:
            path.index = path.index[l:]
            yield path


def difftar2path_iter(diff_tarfile):
    """
    Turn a file-like difftar object into an iterator of ROPaths.

    :param TarFile diff_tarfile: The tarfile object containing the diff.
    :returns: An iterator of ROPaths.
    :rtype: iterator
    """
    tar_iter = iter(diff_tarfile)
    multivol_fileobj = None

    # The next tar_info is stored in this one element list so
    # Multivol_Filelike below can update it.  Any StopIterations will
    # be passed upwards.
    try:
        tarinfo_list = [next(tar_iter)]
    except StopIteration:
        return

    while True:
        # This section relevant when a multivol diff is last in tar
        if not tarinfo_list[0]:
            return
        if multivol_fileobj and not multivol_fileobj.at_end:
            multivol_fileobj.close()  # aborting in middle of multivol
            continue

        index, difftype, multivol = get_index_from_tarinfo(tarinfo_list[0])
        ropath = ROPath(index)
        ropath.init_from_tarinfo(tarinfo_list[0])
        ropath.difftype = difftype
        if difftype == "deleted":
            ropath.type = None
        elif ropath.isreg():
            if multivol:
                multivol_fileobj = Multivol_Filelike(diff_tarfile, tar_iter, tarinfo_list, index)
                ropath.setfileobj(multivol_fileobj)
                yield ropath
                continue  # Multivol_Filelike will reset tarinfo_list
            else:
                ropath.setfileobj(diff_tarfile.extractfile(tarinfo_list[0]))
        yield ropath
        try:
            tarinfo_list[0] = next(tar_iter)
        except StopIteration:
            return


def get_index_from_tarinfo(tarinfo):
    """
    Return (index, difftype, multivol) pair from tarinfo object.

    :param TarInfo tarinfo: The tarinfo object to parse.
    :returns: A tuple of (index, difftype, multivol).
    :rtype: tuple
    :raises PatchDirException: If the diff entry is unrecognized or contains '..'.
    """
    for prefix in [
        "snapshot/",
        "diff/",
        "deleted/",
        "multivol_diff/",
        "multivol_snapshot/",
    ]:
        tiname = util.get_tarinfo_name(tarinfo)
        if tiname.startswith(prefix):
            name = tiname[len(prefix) :]  # strip prefix
            if prefix.startswith("multivol"):
                if prefix == "multivol_diff/":
                    difftype = "diff"
                else:
                    difftype = "snapshot"
                multivol = 1
                name, num_subs = re.subn("(?s)^multivol_(diff|snapshot)/?(.*)/[0-9]+$", "\\2", tiname)
                if num_subs != 1:
                    raise PatchDirException(f"Unrecognized diff entry {tiname}")
            else:
                difftype = prefix[:-1]  # strip trailing /
                name = tiname[len(prefix) :]
                if name.endswith(r"/"):
                    name = name[:-1]  # strip trailing /'s
                multivol = 0
            break
    else:
        raise PatchDirException(f"Unrecognized diff entry {tiname}")
    if name == r"." or name == r"":
        index = ()
    else:
        index = tuple(os.fsencode(name).split(b"/"))
        if b".." in index:
            raise PatchDirException(f"Tar entry {os.fsdecode(tiname)} contains '..'.  Security violation")
    return index, difftype, multivol


class Multivol_Filelike(object):
    """Emulate a file like object from multivols

    Maintains a buffer about the size of a volume.  When it is read()
    to the end, pull in more volumes as desired.

    """

    def __init__(self, tf, tar_iter, tarinfo_list, index):
        """
        Initializer.  tf is TarFile obj, tarinfo is first tarinfo
        """
        self.tf, self.tar_iter = tf, tar_iter
        self.tarinfo_list = tarinfo_list  # must store as list for write access
        self.index = index
        self.buffer = b""
        self.at_end = False

    def read(self, length=-1):
        """
        Read length bytes from file
        """
        if length < 0:
            while self.addtobuffer():
                pass
            real_len = len(self.buffer)
        else:
            while len(self.buffer) < length:
                if not self.addtobuffer():
                    break
            real_len = min(len(self.buffer), length)

        result = self.buffer[:real_len]
        self.buffer = self.buffer[real_len:]
        return result

    def addtobuffer(self):
        """
        Add next chunk to buffer
        """
        if self.at_end:
            return False
        index, difftype, multivol = get_index_from_tarinfo(self.tarinfo_list[0])
        if not multivol or index != self.index:
            # we've moved on
            # the following communicates next tarinfo to difftar2path_iter
            self.at_end = True
            return False

        fp = self.tf.extractfile(self.tarinfo_list[0])
        self.buffer += fp.read()
        fp.close()

        try:
            self.tarinfo_list[0] = next(self.tar_iter)
        except StopIteration:
            self.tarinfo_list[0] = None
            self.at_end = True
            return False
        return True

    def close(self):
        """
        If not at end, read remaining data
        """
        if not self.at_end:
            while True:
                self.buffer = b""
                if not self.addtobuffer():
                    break
        self.at_end = True


class TarFile_FromFileobjs(object):
    """
    Like a dup_tarfile.TarFile iterator, but read from multiple fileobjs
    """

    def __init__(self, fileobj_iter):
        """Make new tarinfo iterator

        fileobj_iter should be an iterator of file objects opened for
        reading.  They will be closed at end of reading.

        """
        self.fileobj_iter = fileobj_iter
        self.dup_tarfile, self.tar_iter = None, None
        self.current_fp = None

    def __iter__(self):  # pylint: disable=non-iterator-returned
        """
        Return self as an iterator.
        """
        return self

    def set_tarfile(self):
        """
        Set dup_tarfile from next file object, or raise StopIteration
        """
        if self.current_fp:
            assert (
                not self.current_fp.close()
            ), "Closing current file pointer returned a truthy value indicating an error"

        self.current_fp = next(self.fileobj_iter)

        self.dup_tarfile = util.make_tarfile("r", self.current_fp)
        self.tar_iter = iter(self.dup_tarfile)

    def __next__(self):
        """
        Return the next tarinfo object.
        """
        if not self.dup_tarfile:
            try:
                self.set_tarfile()
            except StopIteration:
                return
        try:
            return next(self.tar_iter)
        except StopIteration:
            assert not self.dup_tarfile.close(), "self.dup_tarfile failed to close"
            self.set_tarfile()
            return next(self.tar_iter)

    def extractfile(self, tarinfo):
        """
        Return data associated with given tarinfo
        """
        return self.dup_tarfile.extractfile(tarinfo)


def collate_iters(iter_list):
    """Collate iterators by index

    Input is a list of n iterators each of which must iterate elements
    with an index attribute.  The elements must come out in increasing
    order, and the index should be a tuple itself.

    The output is an iterator which yields tuples where all elements
    in the tuple have the same index, and the tuple has n elements in
    it.  If any iterator lacks an element with that index, the tuple
    will have None in that spot.

    """
    # overflow[i] means that iter_list[i] has been exhausted
    # elems[i] is None means that it is time to replenish it.
    iter_num = len(iter_list)
    if iter_num == 2:
        return diffdir.collate2iters(iter_list[0], iter_list[1])
    overflow = [None] * iter_num
    elems = overflow[:]

    def setrorps(overflow, elems):
        """
        Set the overflow and rorps list
        """
        for i in range(iter_num):
            if not overflow[i] and elems[i] is None:
                try:
                    elems[i] = next(iter_list[i])
                except StopIteration:
                    overflow[i] = 1
                    elems[i] = None

    def getleastindex(elems):
        """
        Return the first index in elems, assuming elems isn't empty
        """
        return min([elem.index for elem in [x for x in elems if x]])

    def yield_tuples(iter_num, overflow, elems):
        while True:
            setrorps(overflow, elems)
            if None not in overflow:
                break

            index = getleastindex(elems)
            yieldval = []
            for i in range(iter_num):
                if elems[i] and elems[i].index == index:
                    yieldval.append(elems[i])
                    elems[i] = None
                else:
                    yieldval.append(None)
            yield tuple(yieldval)

    return yield_tuples(iter_num, overflow, elems)


def normalize_ps(patch_sequence):
    """Given an sequence of ROPath deltas, remove blank and unnecessary

    The sequence is assumed to be in patch order (later patches apply
    to earlier ones).  A patch is unnecessary if a later one doesn't
    require it (for instance, any patches before a "delete" are
    unnecessary).

    """
    result_list = []
    i = len(patch_sequence) - 1
    while i >= 0:
        delta = patch_sequence[i]
        if delta is not None:
            # skip blank entries
            result_list.insert(0, delta)
            if delta.difftype != "diff":
                break
        i -= 1
    return result_list


def patch_seq2ropath(patch_seq):
    """
    Apply the patches in patch_seq, return single ropath
    """
    first = patch_seq[0]
    assert first.difftype != "diff", f"First patch in sequence {patch_seq} was a diff"
    if not first.isreg():
        # No need to bother with data if not regular file
        assert len(patch_seq) == 1, f"Patch sequence isn't regular, but has {len(patch_seq)} entries"
        return first.get_ropath()

    current_file = first.open("rb")

    for delta_ropath in patch_seq[1:]:
        assert (
            delta_ropath.difftype == "diff"
        ), f"Expected difftype 'diff' in patch sequence, got {delta_ropath.difftype}"
        try:
            cur_file.fileno()
        except Exception as e:
            """
            librsync insists on a real file object, which we create manually
            by using the duplicity.tempdir to tell us where.

            See https://bugs.launchpad.net/duplicity/+bug/670891 for discussion
            of os.tmpfile() vs tempfile.TemporaryFile() w.r.t. Windows / Posix,
            which is worked around in librsync.PatchedFile() now.
            """
            tempfp = tempfile.TemporaryFile(dir=tempdir.default().dir())
            util.copyfileobj(current_file, tempfp)
            assert (
                not current_file.close()
            ), "Closing temporary current_file returned a truthy value indicating an error"
            tempfp.seek(0)
            current_file = tempfp
        current_file = librsync.PatchedFile(current_file, delta_ropath.open("rb"))
    result = patch_seq[-1].get_ropath()
    result.setfileobj(current_file)
    return result


def integrate_patch_iters(iter_list):
    """
    Combine a list of iterators of ROPath patches.

    The ``iter_list`` should be sorted in patch order, and the elements in
    each iterator need to be ordered by index. The output will be an
    iterator of the final ROPaths in index order.

    :param list iter_list: A list of ROPath iterators.
    :returns: An iterator of integrated ROPaths.
    :rtype: iterator
    """
    collated = collate_iters(iter_list)
    for patch_seq in collated:
        normalized = normalize_ps(patch_seq)
        try:
            final_ropath = patch_seq2ropath(normalized)
            if final_ropath.exists():
                # otherwise final patch was delete
                yield final_ropath
        except Exception as e:
            filename = normalized[-1].get_ropath().get_relative_path()
            log.Warn(
                _("Error '%s' patching %s") % (util.uexc(e), os.fsdecode(filename)),
                log.WarningCode.cannot_process,
                util.escape(filename),
            )


def tarfiles2rop_iter(tarfile_list, restrict_index=()):
    """Integrate tarfiles of diffs into single ROPath iter

    Then filter out all the diffs in that index which don't start with
    the restrict_index.

    """
    diff_iters = [difftar2path_iter(x) for x in tarfile_list]
    if restrict_index:
        # Apply filter before integration
        diff_iters = [filter_path_iter(x, restrict_index) for x in diff_iters]
    return integrate_patch_iters(diff_iters)


def Write_ROPaths(base_path, rop_iter):
    """
    Write out ROPaths in ``rop_iter`` starting at ``base_path``.

    :param Path base_path: The base path where to write the ROPaths.
    :param iterator rop_iter: An iterator of ROPaths to write.
    :returns: 1 if something was actually written, 0 otherwise.
    :rtype: int
    """
    ITR = IterTreeReducer(ROPath_IterWriter, [base_path])
    return_val = 0
    for ropath in rop_iter:
        return_val = 1
        ITR(ropath.index, ropath)
    ITR.Finish()
    base_path.setdata()
    return return_val


class ROPath_IterWriter(ITRBranch):
    """Used in Write_ROPaths above

    We need to use an ITR because we have to update the
    permissions/times of directories after we write the files in them.

    """

    def __init__(self, base_path):
        """
        Set base_path, Path of root of tree
        """
        self.base_path = base_path
        self.dir_diff_ropath = None
        self.dir_new_path = None

    def start_process(self, index, ropath):
        """
        Write ropath.  Only handles the directory case.
        """
        if not ropath.isdir():
            # Base may not be a directory, but rest should
            new_path = self.base_path.new_index(index)
            if index != ():
                parent_path = new_path.get_parent_dir()
                if not parent_path.exists():
                    parent_path.mkdir()
            if ropath.exists():
                if new_path.exists():
                    new_path.deltree()
                ropath.copy(new_path)

        self.dir_new_path = self.base_path.new_index(index)
        if not self.dir_new_path.exists():
            self.dir_new_path.mkdir()
        self.dir_diff_ropath = ropath

    def end_process(self):
        """
        Update information of a directory when leaving it.
        """
        if self.dir_diff_ropath:
            self.dir_diff_ropath.copy_attribs(self.dir_new_path)

    def can_fast_process(self, index, ropath):  # pylint: disable=unused-argument
        """
        Can fast process (no recursion) if ropath isn't a directory.
        """
        log.Info(
            _("Writing %s of type %s") % (os.fsdecode(ropath.get_relative_path()), ropath.type),
            log.InfoCode.patch_file_writing,
            f"{util.escape(ropath.get_relative_path())} {ropath.type}",
        )
        return not ropath.isdir()

    def fast_process(self, index, ropath):
        """
        Write non-directory ropath to destination.
        """
        if ropath.exists():
            ropath.copy(self.base_path.new_index(index))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
# Copyright 2008 Michael Terry
# Copyright 2011 Canonical Ltd
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Log various messages depending on verbosity level.
"""

import logging
import os
import sys

MIN = 0
ERROR = 0
WARNING = 2
NOTICE = 3
INFO = 5
DEBUG = 9
MAX = 9

PREFIX = ""

_logger = None


def DupToLoggerLevel(verb):
    """
    Convert duplicity level to the logging module's system.

    In the logging module's system, higher values are more severe.

    :param int verb: The duplicity verbosity level.
    :returns: The corresponding logging module level.
    :rtype: int
    """
    return MAX - verb + 1


def LoggerToDupLevel(verb):
    """
    Convert logging module level to duplicity's system, where lower is more severe
    """
    return DupToLoggerLevel(verb)


def LevelName(level):
    """
    Return the name of the given verbosity level.

    :param int level: The verbosity level.
    :returns: The name of the level (e.g., "DEBUG", "INFO").
    :rtype: str
    """
    if level >= 9:
        return "DEBUG"
    elif level >= 5:
        return "INFO"
    elif level >= 3:
        return "NOTICE"
    elif level >= 1:
        return "WARNING"
    else:
        return "ERROR"


def Log(s, verb_level, code=1, extra=None, force_print=False, transfer_progress=False):
    """
    Write a message to the log if the verbosity level is high enough.

    :param str s: The message to log.
    :param int verb_level: The verbosity level of the message.
    :param int code: An error or info code.
    :param str extra: Extra information to include in the control line.
    :param bool force_print: Whether to force printing regardless of verbosity.
    :param bool transfer_progress: Whether this is a transfer progress message.
    """
    global _logger
    if not _logger:
        setup()

    if extra:
        controlLine = f"{int(code)} {extra}"
    else:
        controlLine = f"{int(code)}"
    if not s:
        s = ""  # If None is passed, standard logging would render it as 'None'

    if force_print:
        initial_level = _logger.getEffectiveLevel()
        _logger.setLevel(DupToLoggerLevel(MAX))

    # If all the backends kindly gave us unicode, we could enable this next
    # assert line.  As it is, we'll attempt to convert s to unicode if we
    # are handed bytes.  One day we should update the backends.
    # assert isinstance(s, unicode)
    if not isinstance(s, str):
        s = s.decode("utf8", "replace")

    s = PREFIX + s

    _logger.log(
        DupToLoggerLevel(verb_level),
        s,
        extra={
            "levelName": LevelName(verb_level),
            "controlLine": controlLine,
            "transferProgress": transfer_progress,
        },
    )

    if force_print:
        _logger.setLevel(initial_level)


def Debug(s):
    """
    Shortcut used for debug message (verbosity 9).
    """
    Log(s, DEBUG)


class InfoCode(object):
    """
    Enumeration class to hold info code values.
    These values should never change, as frontends rely upon them.
    Don't use 0 or negative numbers.
    """

    generic = 1
    progress = 2
    collection_status = 3
    diff_file_new = 4
    diff_file_changed = 5
    diff_file_deleted = 6
    patch_file_writing = 7
    # patch_file_patching = 8 # 8 isn't used anymore.
    # file_list = 9 # 9 isn't used anymore.  It corresponds to an older syntax for listing files
    file_list = 10
    synchronous_upload_begin = 11
    asynchronous_upload_begin = 12
    synchronous_upload_done = 13
    asynchronous_upload_done = 14
    skipping_socket = 15
    upload_progress = 16


def Info(s, code=InfoCode.generic, extra=None):
    """
    Shortcut used for info messages (verbosity 5).
    """
    Log(s, INFO, code, extra)


def Progress(s, current, total=None):
    """
    Shortcut used for progress messages (verbosity 5).
    """
    if total:
        controlLine = f"{int(current)} {int(total)}"
    else:
        controlLine = f"{int(current)}"
    Log(s, INFO, InfoCode.progress, controlLine)


def Notice(s):
    """
    Shortcut used for notice messages (verbosity 3, the default).
    """
    Log(s, NOTICE)


class WarningCode(object):
    """
    Enumeration class to hold warning code values.
    These values should never change, as frontends rely upon them.
    Don't use 0 or negative numbers.
    """

    generic = 1
    orphaned_sig = 2
    unnecessary_sig = 3
    unmatched_sig = 4
    incomplete_backup = 5
    orphaned_backup = 6
    ftp_ncftp_v320 = 7  # moved from error
    cannot_iterate = 8
    cannot_stat = 9
    cannot_read = 10
    no_sig_for_time = 11
    cannot_process = 12
    process_skipped = 13


def Warn(s, code=WarningCode.generic, extra=None):
    """
    Shortcut used for warning messages (verbosity 2)
    """
    Log(s, WARNING, code, extra)


class ErrorCode(object):
    """
    Enumeration class to hold error code values.
    These values should never change, as frontends rely upon them.
    Don't use 0 or negative numbers.  This code is returned by duplicity
    to indicate which error occurred via both exit code and log.
    """

    generic = 1  # Don't use if possible, please create a new code and use it
    command_line = 2
    hostname_mismatch = 3
    no_manifests = 4
    mismatched_manifests = 5
    unreadable_manifests = 6
    cant_open_filelist = 7
    bad_url = 8
    bad_archive_dir = 9
    deprecated_option = 10
    restore_path_exists = 11
    verify_dir_doesnt_exist = 12
    backup_dir_doesnt_exist = 13
    file_prefix_error = 14
    globbing_error = 15
    redundant_inclusion = 16
    inc_without_sigs = 17
    no_sigs = 18
    restore_path_not_found = 19
    no_restore_files = 20
    mismatched_hash = 21
    unsigned_volume = 22
    user_error = 23
    # boto_old_style = 24 # deprecated
    # boto_lib_too_old = 25 # deprecated
    # boto_calling_format = 26 # deprecated
    ftp_ncftp_missing = 27
    ftp_ncftp_too_old = 28
    # ftp_ncftp_v320 = 29 # moved to warning
    exception = 30
    gpg_failed = 31
    # s3_bucket_not_style = 32 # deprecated
    not_implemented = 33
    get_freespace_failed = 34
    not_enough_freespace = 35
    get_ulimit_failed = 36
    maxopen_too_low = 37
    connection_failed = 38
    restart_file_not_found = 39
    gio_not_available = 40
    source_path_mismatch = 42  # 41 is reserved for par2
    ftps_lftp_missing = 43
    volume_wrong_size = 44
    enryption_mismatch = 45
    pythonoptimize_set = 46

    dpbx_nologin = 47

    bad_request = 48
    s3_kms_no_id = 49

    # 50-> 69 reserved for backend errors
    backend_error = 50
    backend_permission_denied = 51
    backend_not_found = 52
    backend_no_space = 53
    backend_command_error = 54
    backend_code_error = 55
    backend_validation_failed = 56

    # file selection filter mode errors
    redundant_filter = 70
    trailing_filter = 71
    absolute_files_from = 72
    empty_files_from = 73

    # gpg key errors
    bad_sign_key = 80
    bad_encrypt_key = 81
    bad_hidden_encrypt_key = 82

    # Reserve 126 because it is used as an error code for pkexec
    # Reserve 127 because it is used as an error code for pkexec
    # Reserve 255 because it is used as an error code for gksu


def Error(s, code=ErrorCode.generic, extra=None):
    """
    Write error message.
    """
    Log(s, ERROR, code, extra)


def FatalError(s, code=ErrorCode.generic, extra=None):
    """
    Write fatal error message and exit.
    """
    from duplicity import errors

    raise errors.FatalError(s, code, extra)


class OutFilter(logging.Filter):
    """
    Filter that only allows warning or less important messages.
    """

    def filter(self, record):
        return record.msg and record.levelno <= DupToLoggerLevel(WARNING)


class ErrFilter(logging.Filter):
    """
    Filter that only allows messages more important than warnings.
    """

    def filter(self, record):
        return record.msg and record.levelno > DupToLoggerLevel(WARNING)


def setup():
    """
    Initialize logging
    """
    global _logger
    if _logger:
        return

    # OK, now we can start setup
    _logger = logging.getLogger("duplicity")

    # Default verbosity allows notices and above
    setverbosity(NOTICE)

    # stdout and stderr are for different logging levels
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.setFormatter(PrettyProgressFormatter())
    outHandler.addFilter(OutFilter())
    _logger.addHandler(outHandler)

    errHandler = logging.StreamHandler(sys.stderr)
    errHandler.setFormatter(PrettyProgressFormatter())
    errHandler.addFilter(ErrFilter())
    _logger.addHandler(errHandler)


class PrettyProgressFormatter(logging.Formatter):
    """
    Formatter that overwrites previous progress lines on ANSI terminals.
    """

    last_record_was_progress = False

    def __init__(self):
        """
        Initialize the pretty progress formatter.
        """
        # 'message' will be appended by format()
        # Note that we use our own, custom-created 'levelName' instead of the
        # standard 'levelname'.  This is because the standard 'levelname' can
        # be adjusted by any library anywhere in our stack without us knowing.
        # But we control 'levelName'.
        logging.Formatter.__init__(self, "%(message)s")

    def format(self, record):
        """
        Format the given record, overwriting previous progress lines if necessary.
        """
        s = logging.Formatter.format(self, record)

        # So we don't overwrite actual log lines
        if self.last_record_was_progress and record.transferProgress:
            # Go up one line, then erase it
            s = f"\x1b[F\x1b[2K{s}"

        self.last_record_was_progress = record.transferProgress

        return s


class DetailFormatter(logging.Formatter):
    """
    Formatter that creates messages in a syntax somewhat like syslog.
    """

    def __init__(self):
        """
        Initialize the detail formatter.
        """
        # 'message' will be appended by format()
        # Note that we use our own, custom-created 'levelName' instead of the
        # standard 'levelname'.  This is because the standard 'levelname' can
        # be adjusted by any library anywhere in our stack without us knowing.
        # But we control 'levelName'.
        logging.Formatter.__init__(self, "%(asctime)s %(levelName)-6s %(message)s")

    def format(self, record):
        """
        Format the given record in a detail syntax.
        """
        s = logging.Formatter.format(self, record)
        return s


class MachineFormatter(logging.Formatter):
    """
    Formatter that creates messages in a syntax easily consumable by other processes.
    """

    def __init__(self):
        """
        Initialize the machine-readable formatter.
        """
        # 'message' will be appended by format()
        # Note that we use our own, custom-created 'levelName' instead of the
        # standard 'levelname'.  This is because the standard 'levelname' can
        # be adjusted by any library anywhere in our stack without us knowing.
        # But we control 'levelName'.
        logging.Formatter.__init__(self, "%(levelName)s %(controlLine)s")

    def format(self, record):
        """
        Format the given record in a machine-readable syntax.
        """
        s = logging.Formatter.format(self, record)

        # Add user-text hint of 'message' back in, with each line prefixed by a
        # dot, so consumers know it's not part of 'controlLine'
        if record.message:
            s += f"\n{record.message}".replace("\n", "\n. ")

        # Add a newline so consumers know the message is over.
        return f"{s}\n"


class MachineFilter(logging.Filter):
    """
    Filter that only allows levels that are consumable by other processes.
    """

    def filter(self, record):
        """
        Return True if the record has a levelName attribute.
        """
        # We only want to allow records that have our custom level names
        return hasattr(record, "levelName")


def add_fd(fd):
    """
    Add a file descriptor to which to write machine-readable logging.

    :param int fd: The file descriptor.
    """
    global _logger
    handler = logging.StreamHandler(os.fdopen(fd, "w"))
    handler.setFormatter(MachineFormatter())
    handler.addFilter(MachineFilter())
    _logger.addHandler(handler)


def add_file(filename):
    """
    Add a file to which to write machine-readable logging.

    :param str filename: The path to the file.
    """
    global _logger
    handler = logging.FileHandler(filename, encoding="utf8")
    handler.setFormatter(MachineFormatter())
    handler.addFilter(MachineFilter())
    _logger.addHandler(handler)


def add_timestamp():
    """
    Add timestamp to logs written
    """
    global _logger

    # remove all handlers
    for handler in _logger.handlers[:]:
        _logger.removeHandler(handler)

    # stdout and stderr are for different logging levels
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.setFormatter(DetailFormatter())
    outHandler.addFilter(OutFilter())
    _logger.addHandler(outHandler)

    errHandler = logging.StreamHandler(sys.stderr)
    errHandler.setFormatter(DetailFormatter())
    errHandler.addFilter(ErrFilter())
    _logger.addHandler(errHandler)


def setverbosity(verb):
    """
    Set the verbosity level.

    :param int verb: The verbosity level to set.
    """
    global _logger
    _logger.setLevel(DupToLoggerLevel(verb))


def getverbosity():
    """
    Get the verbosity level.
    """
    global _logger
    return LoggerToDupLevel(_logger.getEffectiveLevel())


def shutdown():
    """
    Cleanup and flush loggers
    """
    global _logger
    logging.shutdown()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2013 Germar Reitze
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Par2 recovery file wrapper backend.
"""

import os
import re

from duplicity import backend
from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity.errors import BackendException


class Par2Backend(backend.Backend):
    """This backend wrap around other backends and create Par2 recovery files
    before the file and the Par2 files are transfered with the wrapped backend.

    If a received file is corrupt it will try to repair it on the fly.
    """

    def __init__(self, parsed_url):
        """
        Initialize the Par2 backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        backend.Backend.__init__(self, parsed_url)

        self.parsed_url = parsed_url
        try:
            self.redundancy = config.par2_redundancy
        except AttributeError:
            self.redundancy = 10

        try:
            self.volumes = config.par2_volumes
        except AttributeError:
            self.volumes = 1

        try:
            self.common_options = f"{config.par2_options} -q -q"
        except AttributeError:
            self.common_options = "-q -q"

        self.wrapped_backend = backend.get_backend_object(parsed_url.url_string)

        for attr in [
            "_get",
            "_put",
            "_list",
            "_delete",
            "_delete_list",
            "_query",
            "_query_list",
            "_retry_cleanup",
            "_error_code",
            "_move",
            "_close",
        ]:
            if hasattr(self.wrapped_backend, attr):
                setattr(self, attr, getattr(self, attr[1:]))

        # always declare _delete_list support because _delete queries file
        # list for every call
        self._delete_list = self.delete_list

    def transfer(self, method, source_path, remote_filename):
        """create Par2 files and transfer the given file and the Par2 files
        with the wrapped backend.

        Par2 must run on the real filename or it would restore the
        temp-filename later on. So first of all create a tempdir and symlink
        the soure_path with remote_filename into this.
        """
        par2temp = source_path.get_temp_in_same_dir()
        par2temp.mkdir()
        # par2cmdline since v1.0 does not accept symlink as source anymore
        source_hardlink = par2temp.append(remote_filename)
        source_target = source_path.get_canonical()
        if not os.path.isabs(source_target):
            source_target = os.path.join(os.fsencode(os.getcwd()), source_target)
        os.link(source_target, source_hardlink.get_canonical())
        source_hardlink.setdata()
        # double-check that the mandatory and correctly named source file now exist
        if not source_hardlink.exists():
            par2temp.deltree()
            raise errors.FatalError(
                f"FAILED to create hard link for par2 processing as "
                f"'{os.fsdecode(source_hardlink.get_canonical())}'"
            )

        log.Info("Create Par2 recovery files")
        par2create = (
            f"par2 c -r{int(self.redundancy)} -n{int(self.volumes)} {self.common_options} "
            f'"{os.fsdecode(source_hardlink.get_canonical())}"'
        )
        returncode, out, err = self.subprocess_popen(par2create)

        if returncode:
            log.Warn("Failed to create par2 file with requested options, retrying with -n1")
            par2create = (
                f"par2 c -r{int(self.redundancy)} -n1 {self.common_options} "
                f'"{os.fsdecode(source_hardlink.get_canonical())}"'
            )
            returncode, out, err = self.subprocess_popen(par2create)
            if not returncode:
                log.Warn("Successfully created par2 file with -n1")

        source_hardlink.delete()
        files_to_transfer = []
        if not returncode:
            for file in par2temp.listdir():
                files_to_transfer.append(par2temp.append(file))
        else:
            log.Error(f"FAILED to create par2 file with returncode {int(returncode)}")

        method(source_path, remote_filename)
        for file in files_to_transfer:
            method(file, file.get_filename())

        par2temp.deltree()

    def put(self, local, remote):
        self.transfer(self.wrapped_backend._put, local, remote)

    def move(self, local, remote):
        self.transfer(self.wrapped_backend._move, local, remote)

    def get(self, remote_filename, local_path):
        """transfer remote_filename and the related .par2 file into
        a temp-dir. remote_filename will be renamed into local_path before
        finishing.

        If "par2 verify" detect an error transfer the Par2-volumes into the
        temp-dir and try to repair.
        """

        par2temp = local_path.get_temp_in_same_dir()
        par2temp.mkdir()
        local_path_temp = par2temp.append(remote_filename)

        self.wrapped_backend._get(remote_filename, local_path_temp)

        try:
            par2file = par2temp.append(remote_filename + b".par2")
            self.wrapped_backend._get(par2file.get_filename(), par2file)

            par2verify = (
                f"par2 v {self.common_options} {os.fsdecode(par2file.get_canonical())} "
                f'"{os.fsdecode(local_path_temp.get_canonical())}"'
            )
            returncode, out, err = self.subprocess_popen(par2verify)

            if returncode:
                log.Warn(f"File is corrupt. Try to repair {remote_filename}")
                c = re.compile(f"{remote_filename.decode()}\\.vol[\\d+]*\\.par2")
                par2volumes = [f for f in self.wrapped_backend._list() if c.match(os.fsdecode(f))]

                for filename in par2volumes:
                    file = par2temp.append(filename)
                    self.wrapped_backend._get(filename, file)

                par2repair = (
                    f"par2 r {self.common_options} {os.fsdecode(par2file.get_canonical())} "
                    f'"{os.fsdecode(local_path_temp.get_canonical())}"'
                )
                returncode, out, err = self.subprocess_popen(par2repair)

                if returncode:
                    log.Error(f"Failed to repair {remote_filename}")
                else:
                    log.Warn(f"Repair successful {remote_filename}")
        except BackendException:
            # par2 file not available
            pass
        finally:
            local_path_temp.rename(local_path)
            par2temp.deltree()

    def delete(self, filename):
        """
        delete given filename and its .par2 files
        """
        self.wrapped_backend._delete(filename)

        remote_list = self.unfiltered_list()

        c = re.compile(f"{os.fsdecode(filename)}(?:\\.vol[\\d+]*)?\\.par2")
        for remote_filename in remote_list:
            if c.match(os.fsdecode(remote_filename)):
                self.wrapped_backend._delete(os.fsencode(remote_filename))

    def delete_list(self, filename_list):
        """
        delete given filename_list and all .par2 files that belong to them
        """
        remote_list = self.unfiltered_list()

        for filename in filename_list[:]:
            c = re.compile(f"{os.fsdecode(filename)}(?:\\.vol[\\d+]*)?\\.par2")
            for remote_filename in remote_list:
                if c.match(os.fsdecode(remote_filename)):
                    # insert here to make sure par2 files will be removed first
                    filename_list.insert(0, remote_filename)

        if hasattr(self.wrapped_backend, "_delete_list"):
            return self.wrapped_backend._delete_list(filename_list)
        else:
            for filename in filename_list:
                self.wrapped_backend._delete(filename)

    def list(self):
        """
        Return a list of filenames (byte strings) present in the backend.

        Files ending with ".par2" will be excluded from the list.

        :returns: A filtered list of filenames.
        :rtype: list
        """
        remote_list = self.wrapped_backend._list()

        c = re.compile("(?!.*\\.par2$)")
        filtered_list = []
        for filename in remote_list:
            if c.match(os.fsdecode(filename)):
                filtered_list.append(filename)
        return filtered_list

    def unfiltered_list(self):
        return self.wrapped_backend._list()

    def retry_cleanup(self):
        self.wrapped_backend._retry_cleanup()

    def error_code(self, operation, e):
        return self.wrapped_backend._error_code(operation, e)

    def query(self, filename):
        return self.wrapped_backend._query(filename)

    def query_list(self, filename_list):
        return self.wrapped_backend._query(filename_list)

    def close(self):
        self.wrapped_backend._close()


backend.register_backend_prefix("par2", Par2Backend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
# Copyright 2013 Edgar Soldin
#                 - ssl cert verification, some robustness enhancements
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
WebDAV backend.
"""

import base64
import http.client
import os
import re
import shutil
import sys
import urllib.error
import urllib.parse
import urllib.request
import xml.dom.minidom

import duplicity.backend
from duplicity import config
from duplicity import errors
from duplicity import log
from duplicity import util
from duplicity.errors import BackendException
from duplicity.errors import FatalBackendException


class CustomMethodRequest(urllib.request.Request):
    """
    This request subclass allows explicit specification of
    the HTTP request method. Basic urllib.request.Request class
    chooses GET or POST depending on self.has_data()
    """

    def __init__(self, method, *args, **kwargs):
        self.method = method
        urllib.request.Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self.method


class CustomHTTPSConnection(http.client.HTTPSConnection):
    """
    Custom HTTPS connection handler that supports certificate verification.
    """

    def __init__(self, *args, **kwargs):
        try:
            global socket, ssl
            import socket
            import ssl
        except ImportError:
            raise FatalBackendException(_("Missing socket or ssl python modules."))

        http.client.HTTPSConnection.__init__(self, *args, **kwargs)

        self.cacert_file = config.ssl_cacert_file
        self.cacert_candidates = [
            "~/.duplicity/cacert.pem",
            "~/duplicity_cacert.pem",
            "/etc/duplicity/cacert.pem",
        ]
        # if no cacert file was given search default locations
        if not self.cacert_file:
            for path in self.cacert_candidates:
                path = os.path.expanduser(path)
                if os.path.isfile(path):
                    self.cacert_file = path
                    break

        # check if file is accessible (libssl errors are not very detailed)
        if self.cacert_file and not os.access(self.cacert_file, os.R_OK):
            raise FatalBackendException(_("Cacert database file '%s' is not readable.") % self.cacert_file)

    def connect(self):
        # create new socket
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self.tunnel()

        if config.ssl_no_check_certificate:
            context = ssl._create_unverified_context()
        else:
            context = ssl.create_default_context(
                ssl.Purpose.SERVER_AUTH,
                cafile=self.cacert_file,
                capath=config.ssl_cacert_path,
            )
        self.sock = context.wrap_socket(sock, server_hostname=self.host)

    def request(self, *args, **kwargs):  # pylint: disable=method-hidden
        try:
            return http.client.HTTPSConnection.request(self, *args, **kwargs)
        except ssl.SSLError as e:
            # encapsulate ssl errors
            raise BackendException(f"SSL failed: {util.uexc(e)}", log.ErrorCode.backend_error)


class WebDAVBackend(duplicity.backend.Backend):
    """Backend for accessing a WebDAV repository.

    webdav backend contributed in 2006 by Jesper Zedlitz
    """

    """
    Request just the names.
    """
    listbody = '<?xml version="1.0"?><D:propfind xmlns:D="DAV:"><D:prop><D:resourcetype/></D:prop></D:propfind>'

    """
Connect to remote store using WebDAV Protocol
"""

    def __init__(self, parsed_url):
        """
        Initialize the WebDAV backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        self.headers = {
            "Connection": "keep-alive",
            "Content-Type": "application/octet-stream",
        }
        if config.webdav_headers:
            try:
                self.headers = util.merge_dicts(self.headers, util.csv_args_to_dict(config.webdav_headers))
            except IndexError as e:
                raise errors.FatalError("--webdav-headers value has an odd number of arguments.  Must be paired.")
            except SyntaxError as e:
                raise errors.FatalError("--webdav-headers value has bad syntax.  Check quoting pairs.")
            except Exception as e:
                raise errors.FatalError(f"--webdav-headers value caused error: {e}")

        self.parsed_url = parsed_url
        self.digest_challenge = None
        self.digest_auth_handler = None

        self.username = parsed_url.username
        self.password = self.get_password()
        self.directory = self.sanitize_path(parsed_url.path)

        log.Debug(_("Using WebDAV host %s port %s") % (parsed_url.hostname, parsed_url.port))
        log.Debug(_("Using WebDAV directory %s") % (self.directory,))

        self.conn = None

    def sanitize_path(self, path):
        if path:
            foldpath = re.compile("/+")
            return foldpath.sub("/", f"{path}/")
        else:
            return "/"

    def getText(self, nodelist):
        rc = ""
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
        return rc

    def _retry_cleanup(self):
        self.connect(forced=True)

    def connect(self, forced=False):
        """
        Connect or re-connect to the server, updates self.conn
        # reconnect on errors as a precaution, there are errors e.g.
        # "[Errno 32] Broken pipe" or SSl errors that render the connection unusable
        """
        if not forced and self.conn and self.conn.host == self.parsed_url.hostname:
            return

        log.Debug(_("WebDAV create connection on '%s'") % self.parsed_url.hostname)
        self._close()
        # http schemes needed for redirect urls from servers
        if self.parsed_url.scheme in ["webdav", "http"]:
            self.conn = http.client.HTTPConnection(self.parsed_url.hostname, self.parsed_url.port)
        elif self.parsed_url.scheme in ["webdavs", "https"]:
            self.conn = CustomHTTPSConnection(self.parsed_url.hostname, self.parsed_url.port)
        else:
            raise FatalBackendException(_("WebDAV Unknown URI scheme: %s") % self.parsed_url.scheme)

        # Workaround cpython http.client issue
        # https://github.com/python/cpython/issues/70107
        # PUT may not return a proper error when ran as first request but throw SSL-EOF-Error or hang
        # as a workaround we run an OPTIONS request to make sure that PUT is never the first request
        response = self.request("OPTIONS", self.directory, None, connect=False)
        response.close()

    def _close(self):
        if self.conn:
            self.conn.close()

    def request(self, method, path, data=None, redirected=0, connect=True):
        """
        Wraps the connection.request method to retry once if authentication is
        required
        """

        def munge_headers(headers):
            """
            Munge http auth headers for debug output below
            """
            headers_copy = dict(headers)
            for key in headers_copy:
                if str(key).lower() == "authorization":
                    headers_copy[key] = "**MUNGED**"
            return headers_copy

        self._close()  # or we get previous request's data or exception
        if connect:
            self.connect()

        quoted_path = urllib.parse.quote(path, "/:~")

        if self.digest_challenge is not None:
            self.headers["Authorization"] = self.get_digest_authorization(path)
        elif self.username or self.password:
            self.headers["Authorization"] = self.get_basic_authorization()

        log.Debug(_("WebDAV %s %s request with headers: %s ") % (method, quoted_path, munge_headers(self.headers)))
        log.Debug(_("WebDAV data length: %s ") % sys.getsizeof(data))
        self.conn.request(method, quoted_path, data, self.headers)
        response = self.conn.getresponse()
        log.Debug(_("WebDAV response status %s with reason '%s'.") % (response.status, response.reason))
        # resolve redirects and reset url on listing requests (they usually come before everything else)
        if response.status in [301, 302] and method == "PROPFIND":
            redirect_url = response.getheader("location", None)
            response.close()
            if redirect_url:
                log.Notice(_("WebDAV redirect to: %s ") % urllib.parse.unquote(redirect_url))
                if redirected > 10:
                    raise FatalBackendException(_("WebDAV redirected 10 times. Giving up."))
                self.parsed_url = duplicity.backend.ParsedUrl(redirect_url)
                self.directory = self.sanitize_path(self.parsed_url.path)
                return self.request(method, self.directory, data, redirected + 1)
            else:
                raise FatalBackendException(_("WebDAV missing location header in redirect response."))
        # mainly for digest-auth to recalculate with response values
        elif response.status == 401:
            response.read()
            response.close()
            self.headers["Authorization"] = self.get_authorization(response, quoted_path)
            log.Debug(_("WebDAV retry request with authentification headers."))
            log.Debug(_("WebDAV %s %s request2 with headers: %s ") % (method, quoted_path, munge_headers(self.headers)))
            log.Debug(_("WebDAV data length: %s ") % sys.getsizeof(data))
            self.conn.request(method, quoted_path, data, self.headers)
            response = self.conn.getresponse()
            log.Debug(_("WebDAV response2 status %s with reason '%s'.") % (response.status, response.reason))

        return response

    def get_authorization(self, response, path):
        """
        Fetches the auth header based on the requested method (basic or digest)
        """
        try:
            auth_hdr = response.getheader("www-authenticate", "")
            token, challenge = auth_hdr.split(" ", 1)
        except ValueError:
            return None
        if token.split(",")[0].lower() == "negotiate":
            try:
                return self.get_kerberos_authorization()
            except ImportError:
                log.Warn(_("python-kerberos needed to use kerberos \
                          authorization, falling back to basic auth."))
                return self.get_basic_authorization()
            except Exception as e:
                log.Warn(_("Kerberos authorization failed: %s.\
                          Falling back to basic auth.") % e)
                return self.get_basic_authorization()
        elif token.lower() == "basic":
            return self.get_basic_authorization()
        else:
            self.digest_challenge = self.parse_digest_challenge(challenge)
            return self.get_digest_authorization(path)

    def parse_digest_challenge(self, challenge_string):
        return urllib.request.parse_keqv_list(urllib.request.parse_http_list(challenge_string))

    def get_kerberos_authorization(self):
        import kerberos  # pylint: disable=import-error

        _, ctx = kerberos.authGSSClientInit(f"HTTP@{self.conn.host}")
        kerberos.authGSSClientStep(ctx, "")
        tgt = kerberos.authGSSClientResponse(ctx)
        return f"Negotiate {tgt}"

    def get_basic_authorization(self):
        """
        Returns the basic auth header
        """
        auth_string = f"{self.username}:{self.password}"
        return f"Basic {base64.b64encode(auth_string.encode()).strip().decode()}"

    def get_digest_authorization(self, path):
        """
        Returns the digest auth header
        """
        u = self.parsed_url
        if self.digest_auth_handler is None:
            pw_manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()
            pw_manager.add_password(None, self.conn.host, self.username, self.password)
            self.digest_auth_handler = urllib.request.HTTPDigestAuthHandler(pw_manager)

        # building a dummy request that gets never sent,
        # needed for call to auth_handler.get_authorization
        scheme = u.scheme == "webdavs" and "https" or "http"
        hostname = u.port and f"{u.hostname}:{u.port}" or u.hostname
        dummy_url = f"{scheme}://{hostname}{path}"
        dummy_req = CustomMethodRequest(self.conn._method, dummy_url)
        auth_string = self.digest_auth_handler.get_authorization(dummy_req, self.digest_challenge)
        return f"Digest {auth_string}"

    def _list(self):
        """
        List files in the remote WebDAV folder.

        :returns: A list of filenames.
        :rtype: list
        """
        response = None
        try:
            self.headers["Depth"] = "1"
            response = self.request("PROPFIND", self.directory, self.listbody)
            del self.headers["Depth"]
            # if the target collection does not exist, create it.
            if response.status == 404:
                response.close()  # otherwise next request fails with ResponseNotReady
                self.makedir()
                # just created an empty folder, so return empty
                return []
            elif response.status in [200, 207]:
                document = response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(f"Bad status code {status} reason {reason}.")

            log.Debug(f"{document}")
            dom = xml.dom.minidom.parseString(document)
            result = []
            for href in dom.getElementsByTagNameNS("*", "href"):
                filename = self.taste_href(href)
                if filename:
                    result.append(filename)
            return result
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def makedir(self):
        """
        Make (nested) directories on the server.
        """
        dirs = self.directory.split("/")
        # url causes directory to start with /, but it might be given
        # with or without trailing / (which is required)
        if dirs[-1] == "":
            dirs = dirs[0:-1]
        for i in range(1, len(dirs)):
            d = f"{'/'.join(dirs[0:i + 1])}/"

            self.headers["Depth"] = "1"
            response = self.request("PROPFIND", d)
            del self.headers["Depth"]

            log.Debug(f"Checking existence dir {d}: {int(response.status)}")

            if response.status == 404:
                log.Debug(_("Creating missing directory %s") % d)

                res = self.request("MKCOL", d)
                if res.status != 201:
                    raise BackendException(_("WebDAV MKCOL %s failed: %s %s") % (d, res.status, res.reason))

    def taste_href(self, href):
        """
        Check the given href node and return the filename if it is a duplicity file.

        :param href: The XML node to check.
        :returns: A matching filename, or None if the href did not match.
        :rtype: str or None
        :raises BackendException: If the hostname in the URL does not match.
        """
        raw_filename = self.getText(href.childNodes).strip()
        parsed_url = urllib.parse.urlparse(urllib.parse.unquote(raw_filename))
        filename = parsed_url.path
        log.Debug(_("WebDAV path decoding and translation: " "%s -> %s") % (raw_filename, filename))

        # at least one WebDAV server returns files in the form
        # of full URL:s. this may or may not be
        # according to the standard, but regardless we
        # feel we want to bail out if the hostname
        # does not match until someone has looked into
        # what the WebDAV protocol mandages.
        if parsed_url.hostname is not None and not (parsed_url.hostname == self.parsed_url.hostname):
            m = (
                f"Received filename was in the form of a full url, but the hostname ({parsed_url.hostname}) "
                f"did not match that of the webdav backend url ({self.parsed_url.hostname}) - "
                f"aborting as a conservative safety measure. If this happens to you, please report the problem"
            )
            raise BackendException(m)

        if filename.startswith(self.directory):
            filename = filename.replace(self.directory, "", 1)
            return filename
        else:
            return None

    def _get(self, remote_filename, local_path):
        url = self.directory + os.fsdecode(remote_filename)
        response = None
        try:
            target_file = local_path.open("wb")
            response = self.request("GET", url)
            if response.status == 200:
                shutil.copyfileobj(response, target_file)
                assert not target_file.close(), "target_file failed to close"
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_("WebDAV GET Bad status code %s reason %s.") % (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def _put(self, source_path, remote_filename):
        url = self.directory + os.fsdecode(remote_filename)
        response = None
        try:
            source_file = source_path.open("rb")
            response = self.request("PUT", url, source_file.read())
            # 200 is returned if a file is overwritten during restarting
            if response.status in [200, 201, 204]:
                response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_("WebDAV PUT Bad status code %s reason %s.") % (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def _delete(self, filename):
        url = self.directory + os.fsdecode(filename)
        response = None
        try:
            response = self.request("DELETE", url)
            if response.status in [200, 204]:
                response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_("WebDAV DEL Bad status code %s reason %s.") % (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()


duplicity.backend.register_backend("http", WebDAVBackend)
duplicity.backend.register_backend("https", WebDAVBackend)
duplicity.backend.register_backend("webdav", WebDAVBackend)
duplicity.backend.register_backend("webdavs", WebDAVBackend)
duplicity.backend.uses_netloc.extend(["http", "https", "webdav", "webdavs"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2022 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Main for parse command line, check for consistency, and set config
"""

import copy
import inspect
import sys
from textwrap import dedent

# TODO: Remove duplicity.argparse311 when py310 xgoes EOL
from duplicity import argparse311 as argparse
from duplicity import backend
from duplicity import config
from duplicity import cli_util
from duplicity import gpg
from duplicity import errors
from duplicity import log
from duplicity import util
from duplicity.cli_data import *


class DuplicityHelpFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    """
    A custom help formatter for Duplicity.

    Combines ArgumentDefaults and RawDescription formatters.
    """


def make_wide(formatter, w=120, h=46):
    """
    Return a wider HelpFormatter, if possible.

    See: https://stackoverflow.com/a/5464440

    :param formatter: The formatter class to widen.
    :param int w: The target width.
    :param int h: The target help position.
    :returns: A widened formatter class or the original.
    """
    try:
        kwargs = {"width": w, "max_help_position": h}
        formatter(None, **kwargs)
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        warnings.warn("argparse help formatter failed, falling back.")
        return formatter


def new_parser(**kwargs):
    """
    Return a properly defined overrideable parser.

    :param kwargs: Additional arguments for the ArgumentParser.
    :returns: A new ArgumentParser instance.
    :rtype: argparse.ArgumentParser
    """
    action_help = "positional args:\n"
    for var, meta in DuplicityCommands.__dict__.items():
        if var.startswith("__"):
            continue
        action_str = f"  {var2cmd(var)} {' '.join(meta)}"
        action_help += f"{action_str:48}" f"# duplicity {var2cmd(var)} [options] {' '.join(meta)}"
        action_help += "\n"
    action_help += "\n"

    return argparse.ArgumentParser(
        prog="duplicity",
        argument_default=None,
        formatter_class=make_wide(DuplicityHelpFormatter),
        epilog=action_help + help_footer,
        allow_abbrev=True,
        exit_on_error=False,
        **kwargs,
    )


def harvest_namespace(args):
    """
    Copy all arguments and their values to the config module.

    Does not copy attributes that are 'hidden' (start with an underscore) or
    whose name is the empty string.

    :param args: The parsed arguments namespace.
    """
    for f in [x for x in dir(args) if x and not x.startswith("_")]:
        v = getattr(args, f)
        setattr(config, f, v)


def parse_log_options(arglist):
    """
    Parse the commands and options that need to be handled first.

    Mainly ensures that logging goes to the right place with correct verbosity.
    Everything else is passed on to the main parsers.

    :param list arglist: The list of command-line arguments.
    :returns: A tuple of (args, remainder).
    :rtype: tuple
    """
    # set up parent parser
    parser = new_parser(add_help=False)

    # add logging/version options to the parser
    for opt in sorted(logging_options):
        var = opt2var(opt)
        names = [opt] + OptionAliases.__dict__.get(var, [])
        parser.add_argument(*names, **OptionKwargs[var])

    # process parent args now
    try:
        args, remainder = parser.parse_known_intermixed_args(arglist)
    except (argparse.ArgumentError, argparse.ArgumentTypeError) as e:
        command_line_error(str(e))

    return args, remainder


def parse_cmdline_options(arglist):
    """
    Parse the remaining argument list once all options are defined.

    :param list arglist: The list of command-line arguments.
    :returns: The parsed arguments.
    """
    # interpret logging/version options early
    args, remainder = parse_log_options(arglist)

    # set up parser
    parser = new_parser()

    # add all options to the parser
    for opt in sorted(all_options):
        var = opt2var(opt)
        names = [opt] + OptionAliases.__dict__.get(var, [])
        parser.add_argument(*names, **OptionKwargs[var])

    # parse the options
    try:
        args, remainder = parser.parse_known_intermixed_args(remainder)
    except (argparse.ArgumentError, argparse.ArgumentTypeError) as e:
        command_line_error(str(e))

    # let's test the command and try to assume which action,
    # eventually err out if no valid action could be determined/was given
    if len(remainder) == 2 and remainder[0] not in all_commands:
        if is_path(remainder[0]) and is_url(remainder[1]):
            log.Notice(
                _(
                    "No valid action found. Will imply 'backup' because "
                    "a path source was given and target is a url location."
                )
            )
            remainder.insert(0, "backup")
        elif is_url(remainder[0]) and is_path(remainder[1]):
            log.Notice(
                _(
                    "No valid action found. Will imply 'restore' because "
                    "url source was given and target is a local path."
                )
            )
            remainder.insert(0, "restore")
        else:
            msg = _(
                f"Invalid '{remainder[0]}' action and cannot be implied from the "
                f"given arguments:\n{arglist}\n"
                f"Valid actions are: {all_commands}"
            )
            command_line_error(msg)

    # check for added/removed options
    for opt in remainder:
        if opt.startswith("-"):
            if opt in changed_options:
                command_line_error(dedent(f"""\
                        Option '{opt} was changed in 2.0.0.
                            --file-to-restore to --path-to-restore
                            --do-not-restore-ownership to --no-restore-ownership
                            """))
            elif opt in removed_options:
                removed_commands_string = "\n".join(f"    {c}" for c in sorted(removed_options))
                command_line_error(dedent(f"""\
                        Option '{opt}' was removed in 2.0.0.
                        The following options were deprecated and removed in 2.0.0
                        """) + f"{removed_commands_string}")
            elif opt in removed_backup_options and args.action in ("backup", "full", "incremental"):
                removed_commands_string = "\n".join(f"    {c}" for c in sorted(removed_backup_options))
                command_line_error(dedent(f"""\
                        Option '{opt}' was removed for backup actions in 2.0.0.
                        The following options were deprecated and removed in 2.0.0
                        """) + f"{removed_commands_string}")

    # check for proper action
    if remainder and remainder[0] in all_commands:
        args.action = remainder[0]
    else:
        command_line_error("Missing or invalid explicit or implicit action.")

    # translate aliases to long action
    var = cmd2var(args.action)
    arg_checks = DuplicityCommands.__dict__.get(var, None)
    if not arg_checks:
        for var, aliases in CommandAliases.__dict__.items():
            if var.startswith("__"):
                continue
            if args.action in aliases:
                arg_checks = DuplicityCommands.__dict__.get(var, None)
                args.action = var2cmd(var)
    if not arg_checks:
        command_line_error(f"'{args.action}' is not a valid action or alias.")

    # parse the positionals relating to action
    if len(arg_checks) == len(remainder[1:]):
        for name, val in zip(arg_checks, remainder[1:]):
            func = getattr(cli_util, f"check_{name}")
            setattr(config, name, func(val))
    else:
        command_line_error(
            f"Wrong number of positional args for '{args.action}', got {len(remainder[1:])}\n"
            f"Expected {len(arg_checks)} positionals from {remainder[1:]}."
        )

    # harvest args to config
    harvest_namespace(args)

    return args


def process_command_line(cmdline_list):
    """
    Process the command line and set the configuration.

    :param list cmdline_list: The list of command-line arguments.
    :returns: The action to be performed.
    :rtype: str
    """
    # build initial gpg_profile
    config.gpg_profile = gpg.GPGProfile()

    # parse command line
    args = parse_cmdline_options(cmdline_list)

    # if we get a different gpg-binary from the commandline then redo gpg_profile
    if config.gpg_binary is not None:
        src = copy.deepcopy(config.gpg_profile)
        config.gpg_profile = gpg.GPGProfile(
            passphrase=src.passphrase,
            sign_key=src.sign_key,
            recipients=src.recipients,
            hidden_recipients=src.hidden_recipients,
        )
    else:
        if config.use_gpgsm:
            program = "gpgsm"
        else:
            program = "gpg"
        config.gpg_binary = util.which(program)
    gpg_version = ".".join(map(str, config.gpg_profile.gpg_version))
    log.Info(_(f"GPG binary is {config.gpg_binary}, version {gpg_version}"))
    if config.use_gpgsm and config.gpg_profile.gpg_version < (2, 2, 27):
        raise errors.FatalError(f"Version {gpg_version} of gpgsm is not supported.  Minimum version is 2.2.27")

    # --use-agent is not safe for symmetric encryption.
    # Notify user and let them decide.
    if (
        config.use_agent
        and config.gpg_profile.sign_key is None
        and len(config.gpg_profile.recipients) == 0
        and len(config.gpg_profile.hidden_recipients) == 0
    ):
        log.Warn(
            "Option --use-agent is unsafe with symmetric encryption.\n"
            "Refer to https://gitlab.com/duplicity/duplicity/-/issues/799 for more information."
        )

    # shorten incremental to inc, replace backup with inc
    if config.action in ("incremental", "backup"):
        if config.action == "backup":
            config.implied_inc = True
        config.action = "inc"

    # import all backends and determine which one we use
    backend.import_backends()
    remote_url = config.source_url or config.target_url
    if remote_url:
        config.backend = backend.get_backend(remote_url)
    else:
        config.backend = None

    # determine full clean local path
    local_pathname = config.source_path or config.target_path
    if config.action in ["full", "inc", "restore", "verify"]:
        local_path = path.Path(path.Path(local_pathname).get_canonical())
        if config.action == "restore":
            if (local_path.exists() and not local_path.isemptydir()) or (
                os.getcwd() == os.path.realpath(local_path.name).decode()
            ):
                raise errors.FatalError(
                    _(f"Restore '{local_path.uc_name}' is not empty or is current_dir.\nWill not overwrite."),
                    log.ErrorCode.restore_path_exists,
                )
        elif config.action == "verify":
            if not local_path.exists():
                raise errors.FatalError(
                    _(f"Verify directory {local_path.uc_name} does not exist"),
                    log.ErrorCode.verify_dir_doesnt_exist,
                )
        else:
            assert config.action in ("full", "inc"), "action must be full or inc"
            if not local_path.exists():
                raise errors.FatalError(
                    _(f"Backup source directory {local_path.uc_name} does not exist."),
                    log.ErrorCode.backup_dir_doesnt_exist,
                )

        config.local_path = local_path

    # generate backup name and set up archive dir
    if config.backup_name is None:
        config.backup_name = generate_default_backup_name(remote_url)
    set_archive_dir(expand_archive_dir(config.archive_dir, config.backup_name))

    # count is only used by the remove-* commands
    config.keep_chains = config.count

    # restore selection uses archive-relative paths and is applied while restoring
    if config.action == "restore" and config.select_opts:
        if config.restore_path:
            command_line_error(
                "--path-to-restore cannot be combined with restore file selection options. "
                "Use --path-to-restore for one archive path, or --include/--exclude for pattern-based restore."
            )
        unsupported_restore_selection = {
            "--exclude-if-present",
            "--exclude-other-filesystems",
            "--files-from",
        }
        unsupported = sorted({opt for opt, _arg in config.select_opts if opt in unsupported_restore_selection})
        if unsupported:
            command_line_error(
                f"The following file selection options are not supported for restore: {', '.join(unsupported)}"
            )

    # selection only applies to certain commands
    if config.action in ["full", "inc", "verify"]:
        set_selection()

    # print derived info
    log.Info(_(f"Using archive dir: {config.archive_dir_path.uc_name}"))
    log.Info(_(f"Using backup name: {config.backup_name}"))

    return config.action


if __name__ == "__main__":
    import types

    log.setup()
    action = process_command_line(sys.argv[1:])
    for a, v in sorted(config.__dict__.items()):
        if a.startswith("_") or isinstance(config.__dict__[a], types.ModuleType):
            continue
        print(f"{a} = {v} ({type(config.__dict__[a])})")
    print(f"verbosity: {log.getverbosity()!s}")
